#ifndef LIB_ENTRY_POINTS
#define LIB_ENTRY_POINTS


// ====================================
// C lib
// ====================================

#include <cstdio>

#include <vector>

// ====================================
// Eigen
// ====================================
#include <StdVector>


// ====================================
// Other
// ====================================

#include "./hrbf.h"

// ====================================
// non-extren methods
// ====================================

double * EvalCurvature(HRBF_3D_r3::Vector & grad, HRBF_3D_r3::MatrixDD hes);


// ====================================
// Extern
// ====================================

extern "C" 
{
	typedef std::vector<HRBFVector3, Eigen::aligned_allocator<HRBFVector3>> EvalVector;

	
	

	//=====================================================================================================================
	// r^3
	//=====================================================================================================================
	
	__declspec(dllexport) HRBF_3D_r3 * CreateHRBF_r3()
	{		
		HRBF_3D_r3 * hrbf = new HRBF_3D_r3;
		return hrbf;		
	};

	__declspec(dllexport) void ReleaseHRBF_r3(HRBF_3D_r3 * hrbf)
	{
		delete hrbf;
	};
		
	__declspec(dllexport) void AddPoint_r3(HRBF_3D_r3 * hrbf, double x, double y, double z,
		double nx, double ny, double nz)
	{
		hrbf->AddPointNormals(x, y, z, nx, ny, nz);
	};
	
	__declspec(dllexport) void Run_r3(HRBF_3D_r3 * hrbf, double nv, double ng)
	{
		hrbf->hermite_fit(nv, ng);
	};

	__declspec(dllexport) void Clear_r3(HRBF_3D_r3 * hrbf)
	{
		hrbf->Clear();
	};

	__declspec(dllexport) double Eval_r3(HRBF_3D_r3 * hrbf, double x, double y, double z)
	{
		return hrbf->eval(HRBF_3D_r3::Vector(x, y, z));
	};

	__declspec(dllexport) double * Grad_r3(HRBF_3D_r3 * hrbf, double x, double y, double z)
	{				
		HRBF_3D_r3::Vector v = hrbf->grad(HRBF_3D_r3::Vector(x, y, z));

		double * res = new double[3];
		res[0] = v.x();
		res[1] = v.y();
		res[2] = v.z();

		return res;

	};

	__declspec(dllexport) double * GetCurvature_r3(HRBF_3D_r3 * hrbf, EvalVector * evalPoints, bool incorrectFasterEval)
	{		
		if (evalPoints->size() == 0)
		{
			return NULL;
		}
		
		if (incorrectFasterEval)
		{
			HRBF_3D_r3::Vector grad = hrbf->grad((*evalPoints)[0]);
			HRBF_3D_r3::MatrixDD hes = hrbf->hessian((*evalPoints)[0]);

			for (size_t i = 1; i < evalPoints->size(); i++)
			{
				grad += hrbf->grad((*evalPoints)[i]);
				hes += hrbf->hessian((*evalPoints)[i]);
			}

			grad /= static_cast<double>(evalPoints->size());
			hes /= static_cast<double>(evalPoints->size());

			return EvalCurvature(grad, hes);
		}
		else 
		{
			
			HRBF_3D_r3::Vector grad = hrbf->grad((*evalPoints)[0]);
			HRBF_3D_r3::MatrixDD hes = hrbf->hessian((*evalPoints)[0]);

			double * res = new double[2];
		
			double * tmp = EvalCurvature(grad, hes);
			res[0] = tmp[0];
			res[1] = tmp[1];
			
			delete[] tmp;

			for (size_t i = 1; i < evalPoints->size(); i++)
			{
				grad = hrbf->grad((*evalPoints)[i]);
				hes = hrbf->hessian((*evalPoints)[i]);

				tmp = EvalCurvature(grad, hes);
				res[0] += tmp[0];
				res[1] += tmp[1];
				delete[] tmp;
			}

			res[0] /= static_cast<double>(evalPoints->size());
			res[1] /= static_cast<double>(evalPoints->size());

			return res;
		}

		
	};


	//=====================================================================================================================
	// Exp (r^3) 
	//=====================================================================================================================

	__declspec(dllexport) HRBF_3D_exp_r3 * CreateHRBF_exp_r3(double epsilon)
	{
		HRBF_3D_exp_r3 * hrbf = new HRBF_3D_exp_r3;
		hrbf->SetEpsilon(epsilon);
		return hrbf;
	};

	__declspec(dllexport) void ReleaseHRBF_exp_r3(HRBF_3D_exp_r3 * hrbf)
	{
		delete hrbf;
	};

	__declspec(dllexport) void AddPoint_exp_r3(HRBF_3D_exp_r3 * hrbf, double x, double y, double z,
		double nx, double ny, double nz)
	{
		hrbf->AddPointNormals(x, y, z, nx, ny, nz);
	};

	__declspec(dllexport) void Run_exp_r3(HRBF_3D_exp_r3 * hrbf, double nv, double ng)
	{
		hrbf->hermite_fit(nv, ng);
	};

	__declspec(dllexport) void Clear_exp_r3(HRBF_3D_exp_r3 * hrbf)
	{
		hrbf->Clear();
	};

	__declspec(dllexport) double Eval_exp_r3(HRBF_3D_r3 * hrbf, double x, double y, double z)
	{
		return hrbf->eval(HRBF_3D_r3::Vector(x, y, z));
	};

	__declspec(dllexport) double * Grad_exp_r3(HRBF_3D_r3 * hrbf, double x, double y, double z)
	{
		HRBF_3D_r3::Vector v = hrbf->grad(HRBF_3D_r3::Vector(x, y, z));

		double * res = new double[3];
		res[0] = v.x();
		res[1] = v.y();
		res[2] = v.z();

		return res;
	};

	__declspec(dllexport) double * GetCurvature_exp_r3(HRBF_3D_exp_r3 * hrbf, EvalVector * evalPoints, bool incorrectFasterEval)
	{

		if (evalPoints->size() == 0)
		{
			return NULL;
		}

		if (incorrectFasterEval)
		{
			HRBF_3D_r3::Vector grad = hrbf->grad((*evalPoints)[0]);
			HRBF_3D_r3::MatrixDD hes = hrbf->hessian((*evalPoints)[0]);

			for (size_t i = 1; i < evalPoints->size(); i++)
			{
				grad += hrbf->grad((*evalPoints)[i]);
				hes += hrbf->hessian((*evalPoints)[i]);
			}

			grad /= static_cast<double>(evalPoints->size());
			hes /= static_cast<double>(evalPoints->size());

			return EvalCurvature(grad, hes);
		}
		else
		{
			HRBF_3D_r3::Vector grad = hrbf->grad((*evalPoints)[0]);
			HRBF_3D_r3::MatrixDD hes = hrbf->hessian((*evalPoints)[0]);

			double * res = new double[2];

			double * tmp = EvalCurvature(grad, hes);
			res[0] = tmp[0];
			res[1] = tmp[1];
			delete[] tmp;

			for (size_t i = 1; i < evalPoints->size(); i++)
			{
				grad = hrbf->grad((*evalPoints)[i]);
				hes = hrbf->hessian((*evalPoints)[i]);

				tmp = EvalCurvature(grad, hes);
				res[0] += tmp[0];
				res[1] += tmp[1];
				delete[] tmp;
			}

			res[0] /= static_cast<double>(evalPoints->size());
			res[1] /= static_cast<double>(evalPoints->size());

			return res;
		}

		



		//printf("eval all");

		
	};


	
	//=====================================================================================================================
	// Other
	//=====================================================================================================================

	__declspec(dllexport) EvalVector * CreateEvalVector()
	{
		return new EvalVector();
	}

	__declspec(dllexport) void AddEvalPoint(EvalVector * evalVector, double x, double y, double z)
	{
		HRBFVector3 p = HRBFVector3(x, y, z);
		evalVector->push_back(p);

	}

	__declspec(dllexport) void ClearEvalPoint(EvalVector * evalVector)
	{		
		evalVector->clear();

	}

	__declspec(dllexport) void ReleaseEvalVector(EvalVector * mem)
	{
		delete mem;
		mem = NULL;
	};

	__declspec(dllexport) void ReleaseMemory(double * mem)
	{
		delete[] mem;
		mem = NULL;
	};

	

}


#endif