/* This Source Code Form is subject to the terms of the Mozilla Public License,
* v. 2.0. If a copy of the MPL was not distributed with this file,
* You can obtain one at http://mozilla.org/MPL/2.0/.
*/

/// @file hrbf_core.hpp
/// Original author: Gael Guennebaud - gael.guennebaud@inria.fr - http://www.labri.fr/perso/guenneba/
/// Rodolphe Vaillant - (Fixed the gradient evaluation) - http://www.irit.fr/~Rodolphe.Vaillant
/// Martin Prantl - some other changes to make code more readable

#ifndef HRBF_CORE_H
#define HRBF_CORE_H

// ====================================
// Eigen
// ====================================
#include <LU>
#include <Cholesky>
#include <StdVector>

// ====================================
// C libs
// ====================================
#include <vector>
#include <iostream>

// ====================================
// Other
// ====================================

#include "./hrbf_funcs.h"

// ====================================
// Main code
// ====================================




/// @brief fitting surface on a cloud point and evaluating the implicit surface
/// @tparam Scalar : a base type float, double etc.
/// @tparam Dim    : integer of the dimension of the ambient space 
/// (for a implicit surface == 3)
/// @tparam HRBF     : the class of a the radial basis
/// must implement float Rbf::f(float) float Rbf::df(float) float Rbf::ddf(float)
/// (see hrbf_phi_funcs.h for an example)
/// @note Please go see http://eigen.tuxfamily.org to use matrix and vectors 
/// types of this lib. The documentation is pretty good.
template<typename Scalar, int Dim, typename RBFFunc, typename FuncR>
class HRBF
{
public:
	
	//some typedefs
	typedef Eigen::Matrix<Scalar, Dim, Dim>                       MatrixDD;
	typedef Eigen::Matrix<Scalar, Dim, 1>                         Vector;
	typedef Eigen::Matrix<Scalar, Dim, Eigen::Dynamic>            MatrixDX;
	typedef Eigen::Matrix<Scalar, Eigen::Dynamic, Eigen::Dynamic> MatrixXX;
	typedef Eigen::Matrix<Scalar, Eigen::Dynamic, 1>              VectorX;

	typedef Func_r<Scalar, Dim> R;


	/// Each column represents p_i:  VectorX pi = _centers.col(i);
	MatrixDX  _centers;
	
	/// Vector of scalar values alpha
	VectorX   _alphas;
	
	/// Each column represents beta_i:  VectorX bi = _betas.col(i);
	MatrixDX  _betas;

	/// std library in Eigen must have allocator
	/// https://eigen.tuxfamily.org/dox-devel/group__TopicStlContainers.html

	std::vector<Vector, Eigen::aligned_allocator<Vector>> points;
	std::vector<Vector, Eigen::aligned_allocator<Vector>> normals;

	Scalar epsilon = Scalar(1);


	// --------------------------------------------------------------------------

	void SetEpsilon(Scalar epsilon)
	{
		this->epsilon = epsilon;
	}

	void AddPointNormals(double x, double y, double z,
		double nx, double ny, double nz)
	{
		points.push_back(Eigen::Matrix<Scalar, Dim, 1>(x, y, z));
		normals.push_back(Eigen::Matrix<Scalar, Dim, 1>(nx, ny, nz));

	}

	void Clear()
	{
		points.clear();
		normals.clear();
	}
	

	// --------------------------------------------------------------------------

	/// Compute surface interpolation given a set of points and normals.
	/// This solve the linear system of equation to find alpha scalars and beta
	/// beta vectors stored in '_alphas' and '_betas' attributes.
	void hermite_fit(double nv = 0, double ng = 0)
	{
		assert(points.size() == normals.size());

		size_t nb_points = points.size();		
		size_t nb_size = (Dim + 1) * nb_points;
		


		// copy the node centers
		_centers.resize(Dim, nb_points);		
		for (size_t i = 0; i < nb_points; i++)
		{
			_centers.col(i) = points[i];
		}



		// Assemble the "design" and "value" matrix and vector
		MatrixXX  D(nb_size, nb_size);
		VectorX   fx(nb_size);
		
		bool useRegularization = false;
		if (nv != 0) useRegularization = true;
		if (ng != 0) useRegularization = true;

		//.template syntax - see http://www.comeaucomputing.com/techtalk/templates/

		for (size_t i = 0; i < nb_points; i++)
		{
			const Vector & p = points[i];
			const Vector & n = normals[i];

			size_t io = (Dim + 1) * i;
			fx(io) = 0;	
			fx.segment(io + 1, Dim) = n; //fx(io + 1) = n(0); fx(io + 2) = n(1); fx(io + 3) = n(2);
			
			for (size_t j = 0; j < nb_points; j++)
			{
				size_t jo = (Dim + 1) * j;

				R r(p, _centers.col(j));
								
				if (r.f() == 0)
				{
					D.block(io, jo, Dim + 1, Dim + 1).setZero();
				}
				else
				{									
					//Matrix pattern
					//This pattern is taken as submatrix - we can swap rows / cols and get the same result
					/*
					    r = ||p - p0||

						r.diff = p - p0


						| phi(r)      | phi(r) * r |						
						------------------------------
						| phi(r) * r|     X + Y    |						

						X = tensorProduct(r.diff, r.diff) * (1 / r^2) * (phi(r) - phi(r) / r)
						Y(diagonal) = phi(r) / r

					*/

					Scalar df = RBFFunc::df(epsilon, r); //phi(r)
					Scalar df_l = df / r.f(); //  phi(r) / r

					Vector g = df * r.grad(); //phi(r) * r (=> col vector)

					//f(x) part
					D(io, jo) = RBFFunc::f(epsilon, r);

					D.row(io).segment(jo + 1, Dim) = g.transpose();
					D.col(jo).segment(io + 1, Dim) = g;

					//grad f(x) part
					D.block(io + 1, jo + 1, Dim, Dim) = (r.tensorProduct()) * ((RBFFunc::ddf(epsilon, r) - df_l) / r.f2());
					D.block(io + 1, jo + 1, Dim, Dim).diagonal().array() += df_l;
				}

				if ((useRegularization) && (io == jo))
				{
					D(io, jo) += nv;

					for (int xx = 1; xx < Dim + 1; xx++)
					{
						D(io + xx, jo + xx) += ng;
					}
				}

			}
		}
		
		VectorX x = D.lu().solve(fx);

		Eigen::Map< Eigen::Matrix<Scalar, Dim + 1, Eigen::Dynamic> > mx(x.data(), Dim + 1, nb_points);
		
		_alphas = mx.row(0);
		_betas = mx.bottomRows(Dim);
	}

	// -------------------------------------------------------------------------

	/// Evaluate potential f() at position 'x'
	Scalar eval(const Vector& x) const
	{
		Scalar retVal = 0;
		size_t nb_nodes = _centers.cols();

		for (size_t i = 0; i < nb_nodes; ++i)
		{
			R r(x, _centers.col(i));

			if (r.f() > 0)
			{
				retVal += _alphas(i) * RBFFunc::f(epsilon, r);
				retVal += _betas.col(i).dot(r.diff) * RBFFunc::df(epsilon, r) / r.f();
			}
		}
		return retVal;
	}

	// -------------------------------------------------------------------------

	/// Evaluate gradient nabla f() at position 'x'
	Vector grad(const Vector& x) const
	{
		Vector gradVal = Vector::Zero();
		size_t nb_nodes = _centers.cols();

		for (size_t i = 0; i < nb_nodes; i++)
		{
			
			R r(x, _centers.col(i));

			if (r.f() > 0.00001)
			{								
				Scalar df = RBFFunc::df(epsilon, r);
				Scalar df_l = df / r.f(); //  phi(r) / r	
				
				gradVal += _alphas(i) * df * r.grad();

				gradVal += _betas.col(i) * df_l;

				gradVal += _betas.col(i).dot(r.diff) * (r.diff / r.f2()) * (RBFFunc::ddf(epsilon, r) - df_l);
			}
		}
		return gradVal;
	}

	// --------------------------------------------------------------------------

	Vector UnitVector(int index)
	{
		Vector v(Dim);
		v.setZero();
		v(index) = 1;
		return v;
	}

	Vector grad2(int index, const Vector& x)
	{
		Vector gradVal = Vector::Zero();
		size_t nb_nodes = _centers.cols();

		for (size_t i = 0; i < nb_nodes; i++)
		{
			R r(x, _centers.col(i));

			if (r.f() > 0.00001)
			{

				Scalar b = RBFFunc::ddf(epsilon, r) - RBFFunc::df(epsilon, r) / r.f();
				Scalar b_der = RBFFunc::dddf(epsilon, r) * r.df(index) - (RBFFunc::ddf(epsilon, r) * r.df(index) * r.f() - RBFFunc::df(epsilon, r) * r.df(index)) / r.f2();
				
				Vector a = (r.diff / r.f2()) * b;
				Vector a_der = (UnitVector(index) * b + r.diff * b_der) * r.f2() - r.diff * b * (2 * r.diff(index));
				a_der /= (r.f2() * r.f2());


				

				gradVal += _alphas(i) * (RBFFunc::ddf(epsilon, r) * r.df(index) * r.grad() + RBFFunc::df(epsilon, r) * r.grad2(index));

				gradVal += _betas.col(i) * (RBFFunc::ddf(epsilon, r) * r.df(index) * r.f() - RBFFunc::df(epsilon, r) * r.df(index)) / r.f2();
				
				gradVal += _betas.col(i).dot(UnitVector(index)) * a;
				gradVal += _betas.col(i).dot(r.diff) * a_der;


			}

		}

		return gradVal;

	}
	

	MatrixDD hessian(const Vector& x)
	{		
		MatrixDD hessian(Dim, Dim);
		for (int i = 0; i < Dim; i++)
		{
			hessian.col(i) = grad2(i, x);
		}
		return hessian;
	}


}; // END HermiteRbfReconstruction Class =======================================


typedef HRBF<double, 3, Rbf_func_pow3<double, 3>, Func_r<double, 3> > HRBF_3D_r3;
typedef HRBF<double, 3, Rbf_func_exp_pow3<double, 3>, Func_r<double, 3> > HRBF_3D_exp_r3;
typedef Eigen::Matrix<double, 3, 1>  HRBFVector3;

#endif // HRBF_CORE_HPP__

