/* This Source Code Form is subject to the terms of the Mozilla Public License, 
 * v. 2.0. If a copy of the MPL was not distributed with this file, 
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#ifndef HRBF_FUNCS_H
#define HRBF_FUNCS_H

//==================================================================================
//==================================================================================
//==================================================================================

/*-----------------------------------------------------------
Struct:	Func_r

Represent r function of RBF parametr
RBF(r)
r = ||p - p0||

Provides some functions to make code more read-friendly
Has X, Y, Z constants to obtain df(X), df(Y), df(Z)
It can take arbitrary dimension, so df(0) = df(X) and we can also
use df(4), df(5), df(..)
-------------------------------------------------------------*/
template <typename Scalar, int Dim>
struct Func_r
{
	static const int X = 0;
	static const int Y = 1;
	static const int Z = 2;

	typedef Eigen::Matrix<Scalar, Dim, Dim> MatrixDD;
	typedef Eigen::Matrix<Scalar, Dim, 1>   Vector;

	const Vector & p;
	const Vector & p0;
	const Vector diff;
	const Scalar length;

	inline Func_r(const Vector & p, const Vector & p0) : 
		p(p), p0(p0), diff(p - p0), length(diff.norm()) { }

	inline Scalar f() { return length; }

	inline Scalar f2() { return length * length; }

	inline Scalar df(int index) { return diff(index) / length; }
	
	inline Vector grad() { return diff / length; }



	inline Scalar ddf(int dx, int ddx) 
	{ 		
		Scalar ddf = (diff(dx) * diff(ddx)) / (length * length * length);

		if (dx == ddx)
		{
			return (Scalar(1) / length) - ddf;
		}

		return -ddf;
	}

	inline Vector grad2(int dx) 
	{ 
		Vector v(Dim);		
		for (int i = 0; i < Dim; i++)
		{
			v(i) = ddf(dx, i);
		}

		return v; 
	}

	inline MatrixDD tensorProduct() { return diff * diff.transpose(); }

};

//==================================================================================
//==================================================================================
//==================================================================================
/*
template<class Derived, typename Scalar, int Dim>
struct IRbf_func
{

	static inline Scalar f(const Scalar& x) { return Derived::fx(x); }
	static inline Scalar df(const Scalar& x) { return Derived::dfx(x); }
	static inline Scalar ddf(const Scalar& x) { return Derived::ddfx(x); }

	static inline Scalar f(const Func_r<Scalar, Dim> & r) { return f(r.length); }
	static inline Scalar df(const Func_r<Scalar, Dim> & r) { return df(r.length); }
	static inline Scalar ddf(const Func_r<Scalar, Dim> & r) { return ddf(r.length); }

};
*/





/*-----------------------------------------------------------
struct:	Rbf_func_pow3

Represent RBF function with parameter r (aka Func_r)
RBF(r)
r = ||p - p0||

phi(x) = r^3
-------------------------------------------------------------*/
template<typename Scalar, int Dim>
struct Rbf_func_pow3 
{
	
	static inline Scalar f(Scalar eps, const Scalar& x) { return x*x*x; }
	static inline Scalar df(Scalar eps, const Scalar& x) { return Scalar(3) * x * x; }
	static inline Scalar ddf(Scalar eps, const Scalar& x) { return Scalar(6) * x; }
	static inline Scalar dddf(Scalar eps, const Scalar& x) { return Scalar(6); }
	
	static inline Scalar f(Scalar eps, const Func_r<Scalar, Dim> & r) { return f(eps, r.length); }
	static inline Scalar df(Scalar eps, const Func_r<Scalar, Dim> & r) { return df(eps, r.length); }
	static inline Scalar ddf(Scalar eps, const Func_r<Scalar, Dim> & r) { return ddf(eps, r.length); }
	static inline Scalar dddf(Scalar eps, const Func_r<Scalar, Dim> & r) { return dddf(eps, r.length); }
	
};





template<typename Scalar, int Dim>
struct Rbf_func_exp_pow3 
{
	static inline Scalar f(Scalar eps, const Scalar& x) 
	{ 
		return exp(-eps * x * x * x); 
	}


	static inline Scalar df(Scalar eps, const Scalar& x)
	{ 
		return -3 * eps * x * x * f(eps, x); 
	}

	static inline Scalar ddf(Scalar eps, const Scalar& x)
	{ 
		return 3 * eps * x * f(eps, x) * (3 * eps * x * x * x - 2);
	}

	static inline Scalar dddf(Scalar eps, const Scalar& x)
	{ 
		Scalar x3 = x * x * x;
		Scalar x6 = x3 * x3;
		return -3 * eps * f(eps, x) * (9 * eps * eps * x6 - 18 * eps * x3 + 2);
	}

	static inline Scalar f(Scalar eps, const Func_r<Scalar, Dim> & r) { return f(eps, r.length); }
	static inline Scalar df(Scalar eps, const Func_r<Scalar, Dim> & r) { return df(eps, r.length); }
	static inline Scalar ddf(Scalar eps, const Func_r<Scalar, Dim> & r) { return ddf(eps, r.length); }
	static inline Scalar dddf(Scalar eps, const Func_r<Scalar, Dim> & r) { return dddf(eps, r.length); }

};


#endif //HRBF_PHI_FUNCS_HPP_
