﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Taken from
    /// C# code by Renaud Bédard
    /// http://theinstructionlimit.com/?p=417
    /// 
    /// The algorithm is from the "Fast Poisson Disk Sampling in Arbitrary Dimensions" paper by Robert Bridson
    /// http://www.cs.ubc.ca/~rbridson/docs/bridson-siggraph07-poissondisk.pdf
    /// </summary>
    public class PoisonDiscSampling
    {
        

        private const int DefaultPointsPerIteration = 30;

        private static readonly float SquareRootTwo = (float)Math.Sqrt(2);


        struct Settings
        {
            public Point3D TopLeft, LowerRight, Center;
            public Point3D Dimensions;
            public float RejectionSqDistance;
            public float MinimumDistance;
            public float CellSize;
            public int GridWidth, GridHeight;
        }

        struct State
        {
            public Point3D?[,] Grid;
            public List<Point3D> ActivePoints, Points;
        }

        private Random rnd;
        private State state;
        private Settings settings;


        private int pointsPerIteration;

        public PoisonDiscSampling(Random rnd, float centerX, float centerY, float radius, double minDist)
        {

            this.rnd = rnd;

            this.pointsPerIteration = DefaultPointsPerIteration;

            this.settings = new Settings();
            this.settings.TopLeft = new Point3D(centerX, centerY, 0) - new Point3D(radius, radius, 0);
            this.settings.LowerRight = new Point3D(centerX, centerY, 0) + new Point3D(radius, radius, 0);
            this.settings.Dimensions = this.settings.LowerRight - this.settings.TopLeft;
            this.settings.Center = (this.settings.TopLeft + this.settings.LowerRight) * 0.5f;
            this.settings.CellSize = (float)minDist / SquareRootTwo;
            this.settings.MinimumDistance = (float)minDist;
            this.settings.RejectionSqDistance = radius * radius;
            this.settings.GridWidth = (int)(settings.Dimensions.X / settings.CellSize) + 1;
            this.settings.GridHeight = (int)(settings.Dimensions.Y / settings.CellSize) + 1;



        }

        public int GetID()
        {
            return this.rnd.Next();
        }

        public List<Point3D> Sample()
        {
            this.state = new State();
            this.state.Points = new List<Point3D>();
            this.state.ActivePoints = new List<Point3D>();
            this.state.Grid = new Point3D?[settings.GridWidth, settings.GridHeight];

            this.AddFirstPoint();

            while (this.state.ActivePoints.Count != 0)
            {
                int listIndex = rnd.Next(this.state.ActivePoints.Count);

                Point3D point = this.state.ActivePoints[listIndex];
                bool found = false;

                for (var k = 0; k < this.pointsPerIteration; k++)
                {
                    found |= this.AddNextPoint(point);
                }

                if (!found)
                {
                    this.state.ActivePoints.RemoveAt(listIndex);
                }
            }

            return state.Points;
        }

        private bool AddNextPoint(Point3D point)
        {
            bool found = false;
            Point3D q = this.GenerateRandomAround(point);

            if (q.X >= this.settings.TopLeft.X &&
                q.X < this.settings.LowerRight.X &&
                q.Y > this.settings.TopLeft.Y &&
                q.Y < this.settings.LowerRight.Y &&
                (Point3D.DistanceSquared(this.settings.Center, q) <= this.settings.RejectionSqDistance))
            {
                Point3D qIndex = this.Denormalize(q, this.settings.TopLeft, this.settings.CellSize);
                bool tooClose = false;

                for (int i = (int)Math.Max(0, qIndex.X - 2); i < Math.Min(this.settings.GridWidth, qIndex.X + 3) && !tooClose; i++)
                {
                    for (int j = (int)Math.Max(0, qIndex.Y - 2); j < Math.Min(this.settings.GridHeight, qIndex.Y + 3) && !tooClose; j++)
                    {
                        if (this.state.Grid[i, j].HasValue && (Point3D.Distance(this.state.Grid[i, j].Value, q) < this.settings.MinimumDistance))
                        {
                            tooClose = true;
                        }
                    }
                }


                if (tooClose == false)
                {
                    found = true;
                    this.state.ActivePoints.Add(q);
                    this.state.Points.Add(q);
                    this.state.Grid[(int)qIndex.X, (int)qIndex.Y] = q;
                }

            }
            return found;
        }



        private void AddFirstPoint()
        {

            while (true)
            {
                double d = rnd.NextDouble();
                double xr = this.settings.TopLeft.X + this.settings.Dimensions.X * d;

                d = rnd.NextDouble();
                double yr = this.settings.TopLeft.Y + this.settings.Dimensions.Y * d;

                Point3D p = new Point3D(xr, yr, 0);
                if (Point3D.DistanceSquared(this.settings.Center, p) > this.settings.RejectionSqDistance)
                {
                    continue;
                }

                Point3D index = this.Denormalize(p, settings.TopLeft, settings.CellSize);

                this.state.Grid[(int)index.X, (int)index.Y] = p;

                this.state.ActivePoints.Add(p);
                this.state.Points.Add(p);

                break;
            }
        }

        private Point3D GenerateRandomAround(Point3D center)
        {
            double d = rnd.NextDouble();
            double radius = this.settings.MinimumDistance + this.settings.MinimumDistance * d;

            d = rnd.NextDouble();
            double angle = Math.PI * 2 * d;

            double newX = radius * Math.Sin(angle);
            double newY = radius * Math.Cos(angle);

            return new Point3D(center.X + newX, center.Y + newY, 0);
        }



        private Point3D Denormalize(Point3D point, Point3D origin, double cellSize)
        {
            return new Point3D((int)((point.X - origin.X) / cellSize), (int)((point.Y - origin.Y) / cellSize), 0);
        }


    }

}
