﻿   	
    public class RBFEstimator
    {
		
		//==================================================================
		//Variables
		//==================================================================
       
		InputMesh mesh;
		CornerTable ct;

		string baseFunction;
		string baseFunctionOriginal;
        double[] k1, k2;


        private bool isR3 = true;

        //==================================================================
        //Properties
        //==================================================================
       

        override public double[] K1
        {
            get {return k1;}
        }

        override public double[] K2
        {
            get { return k2; }
        }

        /// <summary>
        /// Base function used for HRBF
        /// For use with C++ Dll select:
        /// a) exp(-epsilon * r^3)
        /// b) r^3
        /// </summary>
		public string BaseFunction
		{
			get 
			{
				return this.baseFunction;
			}
			set 
			{				
				String r = "(((x - x0)^2 + (y - y0)^2 + (z - z0)^2)^0.5)";

				this.baseFunctionOriginal = value;
				this.baseFunction = value;				
				this.baseFunction = this.baseFunction.Replace("r", r);
			}
		}


       
        /// <summary>
        /// Compute r^3 and exp(-epsilon * r^3) with C++ Dll
        /// </summary>
        public bool Poisson { get; set; }

        public double PoissonMinDist { get; set; }

        /// <summary>
        /// Size of neighborhood
        /// </summary>
		public double Neighborhood { get; set; }


        public bool UseFasterEval { get; set; }
     
        public bool EnableMultisample { get; set; }

        List<Point3D> rndPoints = new List<Point3D>();
        int rndPointsID;
        double el;


        //==================================================================
        // Methods
        //==================================================================

        /// <summary>
        /// ctor
        /// set Default base function to r^3
        /// </summary>
        public RBFEstimator() 
		{
			this.BaseFunction = "r^3";
					          
            Neighborhood = 2;		
            Poisson = false;
            PoissonMinDist = 0.3;
            UseFasterEval = false;
            EnableMultisample = true;            
        }

        
		/// <inheritdoc/>
        override public void EstimateCurvatures(InputMesh mesh)
        {
            rndPoints.Clear();

            if (Poisson)
            {
                PoisonDiscSampling ps = new PoisonDiscSampling(Rand, 0, 0, 1.0f, PoissonMinDist);
                rndPointsID = ps.GetID();
                List<Point3D> points = ps.Sample();

                for (int i = 0; i < points.Count; i++)
                {
                    rndPoints.Add(points[i]);
                }
            }
            else
            {
                rndPointsID = Rand.Next();
                for (int i = 0; i < 30; i++)
                {
                    Point3D rp = Utils.RandomInCircle(Rand, 1.0, true);
                    rndPoints.Add(rp);
                }
            }

            Console.WriteLine("Random points: " + rndPoints.Count);

            this.mesh = mesh;

            k1 = new double[mesh.Points.Length];
            k2 = new double[mesh.Points.Length];

            this.ct = new CornerTable(mesh);
            
            this.isR3 = false;
            if (this.baseFunctionOriginal.Equals("r^3"))
            {
                this.isR3 = true;               
            }
            
			this.UseDLL();
                     
        }


       

        private double ComputeDataAABBSize()
        {
            Point3D min = mesh.Points[0];
            Point3D max = mesh.Points[0];
            for (int k = 1; k < mesh.Points.Length; k++)
            {
                if (mesh.Points[k].X < min.X) min.X = mesh.Points[k].X;
                if (mesh.Points[k].Y < min.Y) min.Y = mesh.Points[k].Y;
                if (mesh.Points[k].Z < min.Z) min.Z = mesh.Points[k].Z;

                if (mesh.Points[k].X > max.X) max.X = mesh.Points[k].X;
                if (mesh.Points[k].Y > max.Y) max.Y = mesh.Points[k].Y;
                if (mesh.Points[k].Z > max.Z) max.Z = mesh.Points[k].Z;
            }

            double xSize = Math.Abs(max.X - min.X);
            double ySize = Math.Abs(max.Y - min.Y);
            double zSize = Math.Abs(max.Z - min.Z);

            double aabbSize = xSize;
            if (aabbSize < ySize) aabbSize = ySize;
            if (aabbSize < zSize) aabbSize = zSize;

            return aabbSize;
        }

        private double ComputeEpsilon(double aabbSize, int i, int[] indices)
        {
            Point3D p = mesh.Points[i];

            double maxEl2 = double.MinValue;
            double minEl2 = double.MaxValue;
            for (int j = 0; j < indices.Length; j++)
            {

                double dist2 = (mesh.Points[indices[j]] - p).Abs2();
                if (dist2 > maxEl2) maxEl2 = dist2;
                if (dist2 < minEl2) minEl2 = dist2;

            }
            double avDist = 0.5 * (Math.Sqrt(maxEl2) + Math.Sqrt(minEl2));


            //double epsilonValue = 0.001;

            double epsilonValue = avDist / (aabbSize * aabbSize * aabbSize);
            
            return epsilonValue;
        }

        //===============================================================================
        // Dll version
        //===============================================================================


        #region DLL_calls


        const string DllPath = "HRBFLibrary.dll";

        [DllImport(DllPath, EntryPoint = "CreateHRBF_r3", CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr CreateHRBF_r3_DLL();

        [DllImport(DllPath, EntryPoint = "CreateHRBF_exp_r3", CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr CreateHRBF_exp_r3_DLL(double epsilon);


        [DllImport(DllPath, EntryPoint = "ReleaseHRBF", CallingConvention = CallingConvention.Cdecl)]
        public static extern void ReleaseHRBF_DLL(IntPtr hrbf);

        [DllImport(DllPath, EntryPoint = "AddPoint", CallingConvention = CallingConvention.Cdecl)]
        public static extern void AddPoint_DLL(IntPtr hrbf, double x, double y, double z,
            double nx, double ny, double nz);

        [DllImport(DllPath, EntryPoint = "Run", CallingConvention = CallingConvention.Cdecl)]
        public static extern void Run_DLL(IntPtr hrbf, double nv, double ng);

        [DllImport(DllPath, EntryPoint = "Clear", CallingConvention = CallingConvention.Cdecl)]
        public static extern void Clear_DLL(IntPtr hrbf);

        [DllImport(DllPath, EntryPoint = "Eval", CallingConvention = CallingConvention.Cdecl)]
        public static extern double Eval_DLL(IntPtr hrbf, double x, double y, double z);

        [DllImport(DllPath, EntryPoint = "Grad", CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr Grad_DLL(IntPtr hrbf, double x, double y, double z);

        [DllImport(DllPath, EntryPoint = "GetCurvature", CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr GetCurvature_DLL(IntPtr hrbf, IntPtr evalVector, bool incorrectFasterEval);

        
        [DllImport(DllPath, EntryPoint = "CreateEvalVector", CallingConvention = CallingConvention.Cdecl)]
        public static extern IntPtr CreateEvalVector_DLL();


        [DllImport(DllPath, EntryPoint = "AddEvalPoint", CallingConvention = CallingConvention.Cdecl)]
        public static extern void AddEvalPoint_DLL(IntPtr evalVector, double x, double y, double z);

        [DllImport(DllPath, EntryPoint = "ClearEvalPoint", CallingConvention = CallingConvention.Cdecl)]
        public static extern void ClearEvalPoint_DLL(IntPtr evalVector);

        [DllImport(DllPath, EntryPoint = "ReleaseEvalVector", CallingConvention = CallingConvention.Cdecl)]
        public static extern void ReleaseEvalVector_DLL(IntPtr evalVector);


        [DllImport(DllPath, EntryPoint = "ReleaseMemory", CallingConvention = CallingConvention.Cdecl)]
        public static extern void ReleaseMemory_DLL(IntPtr mem);

        #endregion


        private void UseDLL() 
        {
            double aabbSize = 0;
            if ((EnableMultisample) || (this.isR3 == false))
            {
                aabbSize = ComputeDataAABBSize();
                el = mesh.AverageEL();
            }
            

            Console.WriteLine("HRBF DLL version - " + this.baseFunctionOriginal);
		

            //init HRBF for parallel processing
            IntPtr[] hrbfs = new IntPtr[100];
            for (int i = 0; i < hrbfs.Length; i++)
            {
                hrbfs[i] = IntPtr.Zero;
                if (this.isR3)
                {
                    hrbfs[i] = CreateHRBF_r3_DLL();
                }                
            }

           
            Console.WriteLine("AEL: " + el);
            Console.WriteLine("Neighborhood: " + this.Neighborhood);
            
            Action<int> ac = (int i) =>
            {
                int threadID = Thread.CurrentThread.ManagedThreadId;

                //get all point in vertex K-ring neighborhood (K = this.Neighborhood)
                int[] indices = ct.VVK(i, (int)this.Neighborhood);

                if (indices.Length < 2)
                {
                    return;
                }
                

                IntPtr hrbf = IntPtr.Zero;
                if (this.isR3)
                {
                    if (threadID >= hrbfs.Length)
                    {                    
                        hrbf = CreateHRBF_r3_DLL();
                    }                                     
                    else
                    {
                        hrbf = hrbfs[threadID];
                    }
                }                
                else
                {                   
                    double epsilonValue = ComputeEpsilon(aabbSize, i, indices);

                    hrbf = CreateHRBF_exp_r3_DLL(epsilonValue);
                }
                

                //-------

               
                for (int j = 0; j < indices.Length; j++)
                {
                    int id = indices[j];
                    Point3D p = mesh.Points[id];
                    Point3D n = mesh.Normals[id];

                    AddPoint_DLL(hrbf, p.X, p.Y, p.Z, n.X, n.Y, n.Z);
                    
                }

                Run_DLL(hrbf, 0, 0);
                
               
                double K1, K2;

                if (EnableMultisample)
                {
                    this.ComputeWeightenedCurvature_DLL(hrbf, i, out K1, out K2);
                }
                else
                {
                   
                    IntPtr evalVector = CreateEvalVector_DLL();
                    AddEvalPoint_DLL(evalVector, mesh.Points[i].X, mesh.Points[i].Y, mesh.Points[i].Z);

                    IntPtr ptr = GetCurvature_DLL(hrbf, evalVector, false);
                    

                    double[] result = new double[2];
                    Marshal.Copy(ptr, result, 0, 2);

                    K1 = result[0];
                    K2 = result[1];
                   

                    ReleaseMemory_DLL(ptr);

                    ReleaseEvalVector_DLL(evalVector);    
                }
                

                this.k1[i] = K1;// (Utils.Clamp(-10, 10, K1));
                this.k2[i] = K2;// (Utils.Clamp(-10, 10, K2));

                if (this.isR3 == false)
                {
                    ReleaseMemory_DLL(hrbf);
                }
                else if (threadID >= hrbfs.Length)
                {
                    ReleaseMemory_DLL(hrbf);
                }
                else
                {
                   Clear_DLL(hrbf);                   
                }
            };


            Parallel.For(0, mesh.Points.Length, ac);

            //for (int i = 0; i < mesh.Points.Length; i++) ac(i);                     
        }
     
        private void ComputeWeightenedCurvature_DLL(IntPtr hrbf, int i, out double K1, out double K2)
        {

            IntPtr evalVector = CreateEvalVector_DLL();


            double radius = el;
            //----------------------------------------------------------------
            //Transform random disc to the tangent plane

            Point3D center = mesh.Points[i];
            Point3D normal = mesh.Normals[i];


            double costheta = normal.Z; //Point3D.DotProduct(new Point3D(0, 0, 1), normal);
            Point3D axis = Point3D.CrossProduct(new Point3D(0, 0, 1), normal).Normalize();

            Matrix4x4 rot = Matrix4x4.Rotation(axis, Math.Acos(costheta));

            for (int t = 0; t < rndPoints.Count; t++)
            {
                Point3D rp = rndPoints[t];
                rp *= radius;
                rp = Point3D.TransformCoordinate(rp, rot);
                rp += center;

                AddEvalPoint_DLL(evalVector, rp.X, rp.Y, rp.Z);
            }

            //----------------------------------------------------------------


            IntPtr ptr = GetCurvature_DLL(hrbf, evalVector, UseFasterEval);
                        
            double[] result = new double[2];
            Marshal.Copy(ptr, result, 0, 2);

            K1 = result[0];
            K2 = result[1];
            
           
            ReleaseMemory_DLL(ptr);
            ReleaseEvalVector_DLL(evalVector);
        }


    }

